package de.uni_heidelberg.ise.studentnavigator.model;

import java.io.BufferedInputStream;
import java.io.BufferedReader;
import java.io.DataInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;

import org.json.JSONArray;
import org.json.JSONException;
import org.json.JSONObject;

import com.hdm_i.dm.android.mapsdk.HDMFeature;
import com.hdm_i.dm.android.mapsdk.MapView;

import de.uhd.ifi.se.feature.annotations.java.Feature;
import de.uni_heidelberg.ise.studentnavigator.model.Person.Occupation;
import de.uni_heidelberg.ise.studentnavigator.model.api.DataManager;

@Feature("Routenfindung")
public class DataToFeatureIdMapper {
    private final ArrayList<HDMFeature> allFeatures = new ArrayList<HDMFeature>();
    private final DataManager dataManager;
    private final MapView deepmap;

    private final HashMap<JSONObject, HDMFeature> featureForJsonObject = new HashMap<JSONObject, HDMFeature>();

    private final HashMap<String, HDMFeature> featureForOriginalSerial = new HashMap<String, HDMFeature>();
    private final BufferedInputStream featureIdStream;
    private final HashMap<HDMFeature, JSONObject> jsonObjectForFeature = new HashMap<HDMFeature, JSONObject>();

    private JSONArray jsonPersons;
    private JSONArray jsonRooms;

    private final BufferedInputStream lsfStream;

    private final ArrayList<String> names = new ArrayList<String>();

    public DataToFeatureIdMapper(MapView deepmap, BufferedInputStream featureIdStream, BufferedInputStream lsfStream,
            DataManager dataManager) {
        this.deepmap = deepmap;
        this.featureIdStream = featureIdStream;
        this.lsfStream = lsfStream;
        this.dataManager = dataManager;

        final ArrayList<Long> ids = ReadFeatureIds();
        ReadLSFData();
        LoadFeatures(ids);
        MapFeatureIdOnLSFId();
        CreatePersons();
        CheckDataObjects();
    }

    private void CheckDataObjects() {
        dataManager.getBuildings();
        dataManager.getOffices();
        dataManager.getPersons();
    }

    private void CreatePerson(JSONObject person) {
        String forename;
        try {
            forename = person.getString("forename");
            final String surname = person.getString("surname"), title = person.getString("title"),
                    email = person.getString("mail"), website = person.getString("homepage"),
                    fax = person.getString("fax"), phone = person.getString("phone");
            final Office office = GetOrCreateOffice(person.getJSONObject("raum"));
            if (office != null)
                dataManager.addPerson(forename, surname, title, phone, fax, email, website, Occupation.UNKNOWN,
                        GetOrCreateOrganization("NONE"), office, "");
        } catch (final JSONException e) {
            e.printStackTrace();
        }
    }

    private void CreatePersons() {
        for (int i = 0; i < jsonPersons.length(); ++i)
            try {
                CreatePerson((JSONObject) jsonPersons.get(i));
            } catch (final JSONException e) {
                e.printStackTrace();
            }
    }

    private String GetCorrectSerialNumber(String originalNumber) {
        final String[] parts = originalNumber.split(" / ");
        return parts[0].replace("INF ", "") + "." + parts[1];
    }

    private Building GetOrCreateBuilding(String buildingName) {
        final List<Building> buildings = dataManager.getBuildings();
        for (final Building building : buildings)
            if (building.getName().equals(buildingName))
                return building;
        return dataManager.addBuilding(buildingName, "", "7524");
    }

    private Office GetOrCreateOffice(JSONObject room) {
        try {
            String name = room.getString("name");
            final List<Office> offices = dataManager.getOffices();
            for (final Office office : offices)
                if (office.getName() == name)
                    return office;
            final Building building = GetOrCreateBuilding(room.getString("gebaeude"));
            final Organization organization = GetOrCreateOrganization("NONE");
            final HDMFeature feature = featureForJsonObject.get(room);
            if (feature != null)
                return dataManager.addOffice(name, building, organization, String.valueOf(feature.getFeatureId()));
        } catch (final JSONException e) {
            e.printStackTrace();
        }
        return null;
    }

    private Organization GetOrCreateOrganization(String organizationName) {
        final List<Organization> organizations = dataManager.getOrganizations();
        for (final Organization organization : organizations)
            if (organization.getName() == organizationName)
                return organization;
        return dataManager.addOrganization(organizationName);
    }

    private void LoadFeatures(ArrayList<Long> ids) {
        final java.util.HashSet<String> features = new java.util.HashSet<String>();
        final ArrayList<String> errors = new ArrayList<String>(), originalSerials = new ArrayList<String>();
        for (final Long id : ids)
            try {
                final HDMFeature currentFeature = deepmap.getFeatureWithId(id);
                allFeatures.add(currentFeature);
                final String originalSerial = (String) currentFeature.getAttributes().get("original_serial");
                if (originalSerial != null && originalSerial.contains("@") && !originalSerial.contains("-")
                        && !originalSerial.contains("_")) {
                    featureForOriginalSerial.put(originalSerial.substring(1, originalSerial.indexOf("@")), currentFeature);
                    originalSerials.add(originalSerial);
                }
                features.add(currentFeature.getFeatureType());
            } catch (final Exception e) {
                errors.add(e.toString());
            }
    }

    private void MapFeatureIdOnLSFId() {
        for (int i = 0; i < jsonRooms.length(); ++i)
            try {
                final JSONObject room = (JSONObject) jsonRooms.get(i);
                final String lsfSerial = (String) room.get("name");
                final HDMFeature accordingFeature = featureForOriginalSerial.get(GetCorrectSerialNumber(lsfSerial));
                if (accordingFeature == null)
                    System.out.println(lsfSerial);
                else {
                    featureForJsonObject.put(room, accordingFeature);
                    jsonObjectForFeature.put(accordingFeature, room);
                }
            } catch (final JSONException e) {
                e.printStackTrace();
            }
    }

    private ArrayList<Long> ReadFeatureIds() {
        final DataInputStream dis = new DataInputStream(featureIdStream);
        final BufferedReader reader = new BufferedReader(new InputStreamReader(dis));
        final ArrayList<Long> featureIds = new ArrayList<Long>();
        String line;
        try {while ((line=reader.readLine()) != null)
            featureIds.add(Long.parseLong(line));} catch (final IOException e){e.printStackTrace();}
        return featureIds;
    }

    private void ReadLSFData() {
        final DataInputStream dis = new DataInputStream(lsfStream);
        final BufferedReader reader = new BufferedReader(new InputStreamReader(dis));
        JSONArray persons = new JSONArray();
        String line;
        try {while ((line=reader.readLine()) != null)
            try {
                persons = new JSONArray(line);
            } catch (final JSONException e) {
                e.printStackTrace();
            }} catch (final IOException e){e.printStackTrace();}

        final JSONArray rooms = new JSONArray();
        for (int i = 0; i < persons.length(); ++i)
            try {
                final JSONObject person = (JSONObject) persons.get(i), room = person.getJSONObject("raum");
                names.add(GetCorrectSerialNumber((String) room.get("name")));
                rooms.put(room);
            } catch (final JSONException e) {
                e.printStackTrace();
            }
        jsonRooms = rooms;
        jsonPersons = persons;
    }
}
