import nltk
import string
import re
import os
import pprint
from sklearn.decomposition import TruncatedSVD
from sklearn.feature_extraction.text import TfidfVectorizer
from sklearn.metrics.pairwise import cosine_similarity
from sklearn.preprocessing import Normalizer
from nltk.stem.porter import PorterStemmer
from pathlib import Path
from os import getcwd
# ignore the following folders
ignore_folders = {'.git', '.svn', 'bin', 'icons', 'images'}
# only use files with this ending
used_file_types = {'java', 'xml', 'mf', 'txt','js','json'}


def convert_camel_case(text):
    """
    FIXME: is now also include in jtil and should be used from there
    convert camel case identifiers contained in the text in single terms
    :param text: text string containing camel case identifiers
    :return: text string with camel case identifiers separated in single terms
    """
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1 \2', text)
    return re.sub('([a-z0-9])([A-Z])', r'\1 \2', s1).lower()


def create_token_dict(local_git_repo_path, jira_issue_path, allowed_src_files=None):
    paths = [local_git_repo_path, jira_issue_path]
    to_dict = {}
    for path in paths:
        for subdir, dirs, files in os.walk(path):
            # check if this a folder to be ignored
            ignore_path = False
            for ignore in ignore_folders:
                if ignore in subdir:
                    ignore_path = True
            if not ignore_path:
                for file in files:
                    # check if file should be included (only text files as specified in 'used_file_types' are processed)
                    ignore_file = True
                    for use in used_file_types:
                        if file.endswith(use):
                            ignore_file = False
                    # file_path = subdir + os.path.sep + file
                    file_path = subdir.replace('\\', '/') + '/' + file
                    # a list with allowed files is given make sure to only process these files
                    if allowed_src_files is not None:
                        for af in allowed_src_files:
                            #dbp = Path(file_path).as_posix()
                            #print ("if '%s' endswith() '%s' ?" % (dbp,af))
                            if Path(file_path).as_posix().endswith(af):
                                # we found the file in the list of allowed files and can break
                                ignore_file = False
                                break
                            else:
                                ignore_file = True

                    if not ignore_file:
                        # print("reading file: "+ file_path)
                        # hack for encoding errors: ignore errors silently
                        try:
                            shakes = open(file_path, 'r', encoding="utf8")
                            text = shakes.read()
                        except:
                            text = ''
                            pass
                        # split camel case identifiers in their parts as described in \cite{Ali2011a} and make all
                        # chars lowercase. this is only reasonable for Java based implementations but has no effect
                        # if there are no camel case identifiers
                        lowers = convert_camel_case(text)
                        #lowers = lowers.lower()
                        # remove none word chars
                        # FIXME: Check translate VS sub
                        #no_punctuation = lowers.translate(string.punctuation)
                        no_punctuation = re.sub('[\W_]', ' ', lowers)
                        # create a usable name for requirements (= bugzilla ticket ID) and for code (path in worksapce
                        # and fully qualified class name )
                        to_dict[file_path] = no_punctuation
    return to_dict


def stem_tokens(tokens):
    stemmed = []
    for item in tokens:
        stemmed.append(PorterStemmer().stem(item))
    return stemmed


def tokenize(text):
    tokens = nltk.word_tokenize(text)
    stems = stem_tokens(tokens)
    return stems

def create_index_4files(ftd):
    """
    create a look up table for the processed files with '<Filename> : <INDEX>' pairs
    :param ftd: is the <file : file content token> dictionary file names are used to create the index
    :return: list with index ids for files
    """
    index4files = {}
    i = 0;
    for k, v in ftd.items():
        index4files[i] = k;
        i += 1;
    return index4files


def generate_tr_set(sim_matrix, threshold_p, fi, debug_p=True):
    """
    transform a similarity matrix to a set of trace links in the format:
    dictionary ( entry : (requirement ticket  : list (implementation artifacts))
    :param sim_matrix:
    :param threshold_p:
    :param fi: look up table with entries in the format '<Filename> : <INDEX>' in which index is the position of
    the file in the trace matrix 'sim_matrix'
    :param debug_p:
    :return: dictionary ( entry : (requirement ticket  : list (implementation artifacts)) with trace links
    """
    # FIXME : Convert the trace link data structure to a list in which each element is a list with two elemnet 1: Issue 2: source file
    link_count = 0
    tr_set = {}
    for idx, row in enumerate(sim_matrix):
        for jdx, col in enumerate(row):
            # only create link :
            # if similarity threshold is reached AND
            # if link files are different from each other AND
            # if links goes from requirement (txt) to a implementation artifact (not txt)
            if col >= threshold_p and (fi[idx] != fi[jdx]) and fi[idx].endswith('txt') and 'ISE' in fi[idx] and not fi[jdx].endswith('txt'):
                if debug_p:
                    print(fi[idx] + " -> " + fi[jdx] + " : ", end="")
                    print(col)
                if tr_set.get(fi[idx]) is None:
                    tr_set[fi[idx]] = [fi[jdx]]
                else:
                    # implementation artifacts are used on file level thus a file is only included once
                    if fi[jdx] not in tr_set.get(fi[idx]):
                        tr_set.get(fi[idx]).append(fi[jdx])
                # count all links also if they are duplicates
                link_count += 1
    if debug_p:
        print("Found %d trace link candidates including duplicate links" % link_count)
    return tr_set


def create_IR_links(local_git_repo_path, jira_issue_path, algo='VSM', threshold=0.2, debug=False, allowed_src_files=None):
    token_dict = create_token_dict(local_git_repo_path, jira_issue_path, allowed_src_files)
    if debug:
        print("%d elements in the token dict" % len(token_dict))


    # 1. vsm
    # tf_idf for token set in the dictionary
    tfidf = TfidfVectorizer(max_df=0.5, tokenizer=tokenize, stop_words='english')
    # use the generated token dict and perform vsm based comparison
    # this can take some time
    tfs = tfidf.fit_transform(token_dict.values())
    tm = (tfs * tfs.T).A
    # fixme: use a list and the list index instead of a self generated counter
    fi = create_index_4files(token_dict)

    if algo == 'LSA':
        print("Using '%s' to generate links ..." % algo)
        # 2. lsa
        # based on :
        # * http://stackoverflow.com/questions/18997905/use-latent-semantic-analysis-with-sklearn
        # * http://stackoverflow.com/questions/30590881/python-lsa-with-sklearn
        # vectorizer = TfidfVectorizer(max_df=0.5,stop_words='english')
        # X = vectorizer.fit_transform(token_dict.values())
        # using latent semantic analyze for
        lsa = TruncatedSVD(n_components=100)
        # http://scikit-learn.org/stable/modules/generated/sklearn.decomposition.TruncatedSVD.html
        tm = lsa.fit_transform(tfs)
        # tm_lsa = lsa.transform(tm_lsa)
        tm = Normalizer(copy=False).fit_transform(tm)



    ir_tr_set_raw = generate_tr_set(cosine_similarity(tm), threshold, fi, debug)
    tr_set = [];
    for issue_raw in ir_tr_set_raw:
        # convert the path of the issue txt file to the jira issue id
        issue_id = Path(issue_raw).name
        issue_id = issue_id[:issue_id.find('.txt')]
        for src_file in ir_tr_set_raw[issue_raw]:
            # convert the src file path so that it has the same format as the mylyn interaction logs
            src_file = src_file[src_file.find('StudentNavigator'):]
            #ir_link = [issue_id,src_file,'IR_VSM_LINK','IR_VSM_LINK','IR_VSM_LINK']
            ir_link = [issue_id, src_file]
            if ir_link not in tr_set:
                if debug:
                    print("Appending IR Link Candidate: %s --> %s" % (issue_id, src_file))
                tr_set.append(ir_link)
    return tr_set


if __name__ == '__main__':
    # local source code files from data/git/ folder
    local_git_repo_path = getcwd() + '/data/git/StudentNavigator/'
    # local user stroies from data/ir/ folder
    jira_issue_path = getcwd() + '/data/ir/s_sts/en/'
    # IR token dictionary creation test
    # create and print the token dictionaries used for IR
    # for all user stories and for all source code files
    # by specifying the last 'allowed_src_files' parameter files are restricted
    # to the given list of file endings, if 'None' is used all files are processed
    ir_res_dic = create_token_dict(jira_issue_path, local_git_repo_path, None)
    for di in ir_res_dic:
        print("file '%s' has tokens: \n%s\n" % (di, ir_res_dic[di]))
    print("Created %d token dictionaries" % len(ir_res_dic))
    print()

    # IR link creation test
    # create trace links by using the parameters below and print the resulting links
    t = 0.3
    algo = 'VSM'
    # debug mode puts out additional information about the created links
    # irls = create_IR_links(local_git_repo_path, jira_issue_path, algo, t, True, None)
    irls = create_IR_links(local_git_repo_path, jira_issue_path, algo, t)
    print("Created %d link candidates with %s threshold=%.2f" % (len(irls), algo, t))
    for irl in irls:
        print(irl)
