"""
Interaction based trace link comparision using interaction data stored in Jira

"""
import csv
import string
import os
import errno
import re
from jira import JIRA
from jira import JIRAError
from zipfile import ZipFile
from io import BytesIO
from xml.etree import ElementTree
from datetime import datetime
from datetime import timezone
from pathlib import Path
from googletrans import Translator
from os import getcwd
from ir import create_IR_links
from vcs import match_commit_diffs, get_distinct_files, clone_remote, get_src_files, maintain_file, use_file_extension, \
    use_sub_dir
from scs import add_code_dep_links

date_format = '%Y-%m-%d %H:%M:%S.%f'
# date format for activity tracker created logs
date_format_at_old = '%Y-%m-%dT%H:%M:%S.%f' # without time zone offset
date_format_at = '%Y-%m-%dT%H:%M:%S.%f%z' # with time zone offset
#most_wrong_src_files = ['InfoBoxManager.java','MainActivity.java','StringConstants.java',
#                        'layout_navigation_container.xml']

# issues with data problems
excluded_story_issues = []
# mapping for story issues to their sub-task issues
# this mapping is used as fallback if no working connection to a jira instance is specified
story_2_subtask_map = {'ISE2017-46': ['ISE2017-58', 'ISE2017-59', 'ISE2017-64', 'ISE2017-73', 'ISE2017-102',
                                      'ISE2017-108', 'ISE2017-155', 'ISE2017-156', 'ISE2017-157', 'ISE2017-158'],
                       'ISE2017-84': ['ISE2017-97', 'ISE2017-98', 'ISE2017-99', 'ISE2017-100']}


def fetch_il4_issue(issue, res_folder, parent_issue=None):
    """
    fetch all interaction log attachments from the given issue and store all interaction log entries in
    a single csv file in the given res_folder
    :param issue: the jira issue including its attachments
    :param res_folder: the folder in which the resulting interaction log csv file is stored
    :return: the number of process interaction events (aka interaction log entries )
    """
    print('found %d attachments' % len(issue.fields.attachment))
    # flag to only write teh csv head once for all processed interaction log attachment files from
    # issues in the resulting csv
    csv_header_written = False
    # counter for processed interaction events
    ie_counter = 0
    # set to keep the processed interactions events to avoid duplicate entries in the resulting csv
    processed_ies = set()
    for attachment in issue.fields.attachment:
        # only process the interaction logs recorded in eclipse there could other attachments as well
        if attachment.filename == 'TRACE_LINK_CREATION_ITS.xml.zip':
            print("Processing : '{filename}' of size: {size} byte".format(
                filename=attachment.filename, size=attachment.size))
            # read content of the interaction log zip
            with ZipFile(BytesIO(attachment.get())) as il_zip:
                for filename in il_zip.namelist():
                    with il_zip.open(filename) as il_xml_file:
                        il_xml = ElementTree.fromstring(il_xml_file.read().decode("utf-8"))
                        ies = il_xml.findall('InteractionEvent')
                        c_ies = len(ies)
                        print('found %d interaction log entries' % c_ies)
                        for ie in ies:
                            # some interaction log entries are missing CreationCount and NumEvents
                            # since they are required for csv writing add them if they are missing
                            if 'CreationCount' not in ie.attrib:
                                ie.attrib['CreationCount'] = 0
                            if 'NumEvents' not in ie.attrib:
                                ie.attrib['NumEvents'] = 0
                            #for k in sorted(ie.attrib):
                            # convert the mylyn interaction log style file path to a file system like format
                            ie.attrib['StructureHandle'] = convert_il_log_res_path(ie.attrib['StructureHandle'])
                            # write each InteractionEvent to a line in the csv

                            # if parent exists store all interactions in a single csv file
                            # this is required to aggregate all sub-task interaction logs to a single story issue
                            csv_filename = issue.key
                            if parent_issue is not None:
                                csv_filename = parent_issue.key
                                # in the aggregate to story case of the csv file already exists
                                # writing of header has already been performed
                                if Path(res_folder+csv_filename+'.csv').is_file():
                                    csv_header_written = True

                            with(open(res_folder+csv_filename+'.csv','a', newline='')) as csvFile:
                                wr = csv.DictWriter(csvFile, fieldnames=sorted(ie.attrib), quoting=csv.QUOTE_ALL,
                                                    delimiter=';')
                                # make sure to only write the csv header file once
                                if not csv_header_written:
                                    wr.writeheader()
                                    csv_header_written = True
                                # avoid duplicates only write an interaction event to the csv if there was no other
                                # interaction event with the exact same values
                                if not ie_contained(processed_ies, ie):
                                    wr.writerow(ie)
                                    processed_ies.add(ie)
                                else:
                                    # reduce the counter of processed interaction logs since
                                    # interaction event was not processed
                                    c_ies -= 1
                        ie_counter += c_ies
        else:
            print ('found other attachment %s' % attachment.filename)
    return ie_counter


def convert_il_log_res_path(il_path):
    """
    convert the file path as used in mylyn to a relative files system like path
    :param il_path:
    :return:
    """
    il_path = il_path.replace('{', '.')
    count = il_path.count('.') - 1
    il_path = il_path.replace('.', '/', count)
    il_path = il_path.replace('<', '/')
    il_path = il_path.replace('&lt;', '/')
    il_path = il_path.replace('\\', '')
    il_path = il_path.replace('//', '/')
    # remove the method and attribute information in the file path
    il_path = il_path.split('[')[0]
    # remove first char if char is one of string.punctuation
    for c in string.punctuation:
        if il_path.startswith(c):
            il_path = il_path[1:]
    return il_path


def ie_contained(ies, ie_comp):
    """
    check if and entry with the exact same values of the interaction event 'ie_comp' is contained in the list of
    interaction events 'ies'
    :param ies: list of interaction events
    :param ie_comp: interaction event for which it is checked if it is contained
    :return: True if interaction event 'ie_comp' is contained in the list of interaction eevents 'ies' False if not
    """
    contained = False
    for ie in ies:
        for k in sorted(ie.attrib):
            #print("%s==%s" %(ie.attrib[k],ie_comp.attrib[k]))
            if ie.attrib[k] == ie_comp.attrib[k]:
                contained = True
            else:
                contained = False
                break
        if contained:
            return contained
    return contained


def print_users(issue):
    assignee = issue.fields.assignee
    if assignee is None:
        assignee = 'None'
    print("creator: '%s' assignee: '%s'" % (issue.fields.creator.name, assignee))


def get_changelog4issue(jira, issue):
    print("getting change logs 4 %s:%s" % (issue.fields.issuetype.name,issue.key))
    issue = jira.issue(issue.key, expand='changelog')
    print_users(issue)
    changelog = issue.changelog
    for history in changelog.histories:
        for item in history.items:
            #if item.field == 'status':
            print("Date: '%s' changed '%s' from: '%s' to: '%s'" % (history.created, item.field, item.fromString, item.toString))
    return issue


def contains_issue(issues, issue):
    for i in issues:
        if i.key == issue.key:
            return True
    return False


def gen_gs_quest(data_dir, issues):
    print("generating GS questionnaire")
    create_dirs(data_dir)
    src_set = get_src_files()
    for i in issues:
        user = i.fields.assignee
        if user is None:
            user = i.fields.creator.name
        else:
            user = i.fields.assignee.name
        for sf in sorted(src_set):
            # write trace link candidate to a line in the csv
            ucsv_exits = os.path.isfile(data_dir+user+'.csv')
            with(open(data_dir+user+'.csv', 'a', newline='')) as csvFile:
                q_csvw = csv.writer(csvFile, quoting=csv.QUOTE_ALL, delimiter=';')
                if not ucsv_exits:
                    q_csvw.writerow(['Issue', 'File', 'Link Correct'])
                q_csvw.writerow([i.key, sf])

            #print("%s;%s;%s" % (user, i.key, sf))
    message = "Story Issues: ";
    for i in issues:
        message += i.key + " "
    print("Created GS questionnaire for %d %s, containing %d link candidates."
          % (len(issues), message, (len(issues) * len(src_set))))


def gen_il_logs_raw(jira, data_dir, il_issues):
    create_dirs(data_dir)
    s_issues_q = []
    print("writing interaction logs to '%s'" % data_dir)
    aie_count = 0
    for issue in il_issues:
        # fetch all data for the issue including attachments
        issue = jira.issue(issue.key)
        if issue.fields.attachment is not None:
            if len(issue.fields.attachment) > 0:
                print("fetching attachments for %s" % issue)
                iie_count = fetch_il4_issue(issue, data_dir)
                print("processed %d interaction events for issue %s:%s" %
                      (iie_count, issue.fields.issuetype.name, issue))
                if issue.fields.issuetype.name == 'Story' and not contains_issue(s_issues_q, issue):
                    s_issues_q.append(issue)
                get_changelog4issue(jira, issue)
                # information about the parent issue which is relevant for mapping interactions from sub tasks to their
                # parent isszues
                if hasattr(issue.fields, 'parent'):
                    parent = issue.fields.parent
                    print("Parent issue is %s:%s" % (parent.fields.issuetype.name, parent))
                    parent = get_changelog4issue(jira, parent)
                    if parent.fields.issuetype.name == 'Story' and not contains_issue(s_issues_q, parent):
                        s_issues_q.append(parent)
                aie_count += iie_count
    print("processed %d interaction events overall" % aie_count)
    return s_issues_q


def gen_il_logs_stories(jira, data_dir, il_issues):
    print("")
    print("")
    print("-- gen_il_logs_stories --")
    create_dirs(data_dir)
    for issue in il_issues:
        if issue.fields.issuetype.name == 'Story':
            # XXX: if the story issues themselfs have interaction logs attached
            # these should be used as well!
            print("Processing '%s : %s'" % (issue.key, issue.fields.issuetype.name))
            if not issue.fields.subtasks:
                print("No Sub-Tasks found for Issue '%s : %s' " % (issue.key, issue.fields.issuetype.name))
            else:
                for st_issue in issue.fields.subtasks:
                    print("Processing IL '%s : %s'" % (st_issue.key, st_issue.fields.issuetype.name))
                    # refetching of all issue data seems to be necessary
                    fetch_il4_issue(jira.issue(st_issue.key), data_dir, issue)
            if not issue.fields.issuelinks:
                print("No Links found for Issue '%s : %s' " % (issue.key, issue.fields.issuetype.name))
            else:
                for l_issue in issue.fields.issuelinks:
                    # XXX: l_issue is only a reference object here
                    # if required data of the issue has to be fetched first
                    print("TODO: Processing IL of Linked issue '%s'?" % l_issue)
        else:
            print("'%s : %s' is NOT a Story!" % (issue.key, issue.fields.issuetype.name))


def convert_il_logs_stories(read_data_dir, write_data_dir):
    """
    read the raw log data, convert it to comparable format and filter it according to event type, duration and frequency
    :param read_data_dir:
    :param write_data_dir:
    :return:
    """

    actual_head_files = get_src_files()
    create_dirs(write_data_dir)
    for path, subdirs, files in os.walk(read_data_dir):
        for file in files:
            print("")
            print("%s" % path)
            print("%s" % file)
            with open(path+file, newline='') as csvfile:
                il_rr = csv.reader(csvfile, delimiter=';', quotechar='"')
                ilil = []
                for il_row in il_rr:
                    #print("il_row: %s" % il_row[9])
                    # only include src files in matching sub dirs and with matching file extensions
                    include_ile = False
                    for fe in use_file_extension:
                        for sd in use_sub_dir:
                            # il_row[9] is the file in the interaction log entry
                            # maintain  the file names contained in the links so that file paths are converted
                            # to only be in the actual file structure of the git head revision
                            if il_row[9].endswith(fe) and il_row[9] not in actual_head_files:
                                il_row[9] = maintain_file(il_row[9], actual_head_files)
                            if il_row[9].endswith(fe) and il_row[9].find(sd) != -1:
                                include_ile = True
                                break
                    # only used allowed sub path and file extensions
                    # filter by the properties of each interaction log entry (= line in the csv)
                    # il_row[4] = event type = selection|edit
                    # il_row[3] = level of interest = frequency of interactions
                    # il_row[2] = End Date
                    # il_row[8] = Start Date

                    # il duration end - start date = il_row[2] - il_row[8]
                    try:
                        il_duration = datetime.strptime(il_row[2][:-4], date_format) - datetime.strptime(il_row[8][:-4], date_format)
                        #print("dur: %s" % il_duration)
                    except: #ValueError:
                        pass
                    if include_ile:
                    # if include_ile and il_row[4] == 'selection':
                    # if include_ile and il_duration.total_seconds() > 10.0:
                    #if include_ile and float(il_row[3]) >= 10.0:
                    # if include_ile and float(il_row[3]) > 9.0:
                    # if include_ile and il_duration.total_seconds() > 300.0: # and float(il_row[6]) > 100.0:

                        ile = [str(file)[:-4], il_row[9]]
                        if ile not in ilil:
                            ilil.append(ile)
                            print("%s;%s" % (ile[0],ile[1]))
                            # TODO: write to csv
                            res_exists = os.path.isfile(write_data_dir+file)
                            # all resulting il links in a single csv
                            #res_exists = os.path.isfile(write_data_dir + 'all_il.csv')
                            with(open(write_data_dir+file, 'a', newline='')) as res_csvfile:
                                res_csv_w = csv.writer(res_csvfile, quoting=csv.QUOTE_ALL, delimiter=';')
                                # write the csv header
                                if not res_exists:
                                    res_csv_w.writerow(['Issue', 'File'])
                                res_csv_w.writerow(ile)


def fetch_stories(jira, s_issues):
    """
    fetch the text content of stories and their sub task form jira and translate german texts to english
    :param jira:
    :param data_dir:
    :param s_issues:
    :return:
    """
    print()
    print("-- fetch_stories --")
    #create_dirs(data_dir)
    translator = Translator()
    for s_issue in s_issues:
        if s_issue.fields.issuetype.name == 'Story':
            # list to store all texts of the story and its sub tasks
            s_issue_texts = []
            print("Processing '%s : %s'" % (s_issue.key, s_issue.fields.issuetype.name))
            print("%s | %s | %s" % (s_issue.key, s_issue.fields.summary, s_issue.fields.description))
            s_issue_texts.append(preprocess(s_issue.fields.summary))
            if s_issue.fields.description is not None:
                s_issue_texts.append(preprocess(s_issue.fields.description))
            if not s_issue.fields.subtasks:
                print("No Sub-Tasks found for Issue '%s : %s' " % (s_issue.key, s_issue.fields.issuetype.name))
            #else:
                for st_issue in s_issue.fields.subtasks:
                    print("Processing '%s : %s'" % (st_issue.key, st_issue.fields.issuetype.name))
                    # refetching of all issue data seems to be necessary
                    st_issue = jira.issue(st_issue.key)
                    print("%s | %s | %s" % (st_issue.key, st_issue.fields.summary, st_issue.fields.description))
                    s_issue_texts.append(preprocess(st_issue.fields.summary))
                    if st_issue.fields.description is not None:
                        s_issue_texts.append(preprocess(st_issue.fields.description))
            # translate the texts and store the translate content to txt files
            ts = translator.translate(s_issue_texts, dest='en', src='de')
            # write the translated text content to files for each story
            create_dirs('data/ir/s/en/')
            si_en_wf = open('data/ir/s/en/'+ s_issue.key + '.txt', 'w+', -1, 'UTF-8')
            for t in ts:
                print("'%s' --> '%s'" % (t.origin, t.text))
                si_en_wf.write(t.text +'\n')
            si_en_wf.close()

            # write the original de text values as well
            create_dirs('data/ir/s/de/')
            si_de_wf = open('data/ir/s/de/' + s_issue.key + '.txt', 'w+', -1, 'UTF-8')
            for sit in s_issue_texts:
                si_de_wf.write(sit + '\n')
            si_de_wf.close()

        else:
            print("'%s : %s' is NOT a Story!" % (s_issue.key, s_issue.fields.issuetype.name))


def get_stories_st_mapping(s_issues, print_info=False):
    stories_sts = {}
    for s_issue in s_issues:
        if s_issue.fields.issuetype.name == 'Story':
            # list to store all texts of the story and its sub tasks
            if print_info:
                print("Processing '%s : %s'" % (s_issue.key, s_issue.fields.issuetype.name))
            #stories_sts[s_issue.key] = s_issue.fields.subtasks
            stories_sts[s_issue.key] = []
            if s_issue.fields.subtasks:
                for st_issue in s_issue.fields.subtasks:
                    if print_info:
                        print("Processing '%s : %s'" % (st_issue.key, st_issue.fields.issuetype.name))
                    stories_sts[s_issue.key].append(st_issue.key)
                    #XXX: check if it is reasonable to also process links from sub tasks as well

        else:
            print("'%s : %s' is NOT a Story!" % (s_issue.key, s_issue.fields.issuetype.name))
    return stories_sts


def create_dirs(_dir):
    try:
        if not os.path.exists(_dir):
            os.makedirs(_dir)
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise


def read_links_from_csv(links_csv_path):
    #print("reading: %s" % links_csv_path)
    ret = []
    with open(links_csv_path, 'r') as f:
        # skip the header line of the csv file
        next(f)
        reader = csv.reader(f, delimiter=';')
        ret = list(reader)
    return ret


def aggregate_links(data_dir):
    links = []
    c = 0;
    for root, dirs, files in os.walk(data_dir):
        for file in files:
            if file.endswith('.csv'):
                links.extend(read_links_from_csv(os.path.join(root, file)))
                c = c + 1
    print("%d link candidates have been aggregated from %d csv files contained in '%s'"
          % (len(links), c, data_dir))
    return links


def limit_links(links, limits=['yes']):
    ret = []
    for link in links:
        if link[0] not in excluded_story_issues:
            for limit in limits:
                if str(link[2]).lower() == limit:
                    if link not in ret:
                        ret.append(link)
    print("limited %d links to %d links" % (len(links), len(ret)))
    return ret


def get_unique_issue_id(links):
    issue_ids = []
    for link in links:
        if link[0] not in issue_ids:
            issue_ids.append(link[0])
    return issue_ids


def limit_links2_issues(issues, links, actual_src_files):
    """
    limit the interaction trace links to a given list of jira issue ids
    :param issues: list with jira issue ids
    :param links: trace link set
    :return: the reduced link set
    """
    ret = []
    #srcs = get_src_files()
    for link in links:
        if link[0] in issues and link[1] in actual_src_files and link not in ret:
                #and not link[1].endswith(tuple(most_wrong_src_files)):
                ret.append(link)
    print("limited %d links to %d links" % (len(links), len(ret)))
    return ret


def rem_links_endswith(links, end='xml'):
    for link in links:
        if str(link[1]).lower().endswith(end):
            links = list(filter(lambda a: a != link, links))
    return links


def link_contained(issue_id, src_file, links):
    for link in links:
        if link[0] == issue_id and link[1] == src_file:
            return True
    return False


def preprocess(text):
    """
    convert camel case identifiers contained in the text in single terms, convert to lowercase, remove punctuation
    :param text: text string containing camel case identifiers
    :return: text string with camel case identifiers separated in single terms
    """
    s0 = re.sub('(.)([A-Z][a-z]+)', r'\1 \2', text)
    s1 = re.sub('([a-z0-9])([A-Z])', r'\1 \2', s0).lower()
    s2 = re.sub('[\W_]', ' ', s1)
    return re.sub(' +', ' ', s2)


def get_all(story_subtask_map, issue_id):
    """
    return a list with the story id and all sub-task ids belonging to this story for the given issue_id
    :param story_sub-task_map:
    :param issue_id:
    :return:
    """
    ret = []
    for story in story_subtask_map:
        if story == issue_id:
            ret.append(story)
            ret.extend(story_subtask_map[story])
            return ret
        for sub_task in story_subtask_map[story]:
            if sub_task == issue_id:
                ret.append(story)
                ret.extend(story_subtask_map[story])
                return ret


def trans_src_files2stories(story_subtask_map, files_per_issue, files_in_git_actual=None, print_info=False):
    files_per_story = {}
    for story in sorted(story_subtask_map):
        # get the direct files of the story
        if story in files_per_issue:
            files_per_story[story] = sorted(files_per_issue[story])
            if print_info: print("Processing story %s" % story)
        else:
            files_per_story[story] = []
            if print_info: print("Processing story %s with NO files on its own" % story)
        # get the sub task files
        for subtask in story_subtask_map[story]:
            if subtask in files_per_issue:
                if print_info: print("  Processing files for subtask %s" % subtask)
                files_per_story[story].extend(files_per_issue[subtask])
            else:
                if print_info: print("  No files for subtask %s in files_per_issue" % subtask)
        # remove duplicate files resulting from combining the files of different subtaks
        files_per_story[story] = sorted(list(set(files_per_story[story])))
        # remove files which are not in the actual git revision
        if files_in_git_actual is not None:
            for fps in files_per_story[story]:
                if fps not in files_in_git_actual:
                    # remove all not matching files
                    files_per_story[story] = list(filter(lambda a: a != fps, files_per_story[story]))
    return files_per_story


def reduce_links(links, files_per_issue, mode='story'):
    # simple approach: remove all links with files which are not part of the git files
    if mode == 'simple':
        u_src_files_git = get_distinct_files(files_per_issue)
        for link in links:
            if link[1] not in u_src_files_git:
                links = list(filter(lambda a: a != link, links))
                print("REM: %s:%s" % (link[0], link[1]))
    # 2nd remove links which are not part of the git diffs
    elif mode == 'story':
        for link in links:
            # check if there is a diff (list of files) for the issue in git
            if link[0] not in files_per_issue:
                # if not remove all links which contain the issue not contained in git
                # XXX: sub task to story aggregation!
                links = list(filter(lambda a: a != link, links))
                print("REM: %s:%s" % (link[0], link[1]))
            else:
                # if there is a list of files for the issue check if the file of the link is contained and if not remove
                if link[1] not in files_per_issue[link[0]]:
                    links = list(filter(lambda a: a != link, links))
                    print("REM: %s:%s" % (link[0], link[1]))
    else:
        print("Mode '%s' does NOT exist" % mode)
    return links


def get_distinct_files_links(links):
    files = []
    for link in links:
        if link[1] not in files:
            files.append(link[1])
    return sorted(files)


def print_distinct_files(links):
    for file in sorted(get_distinct_files_links(links)):
        print(file)


def print_list(es):
    for e in sorted(es):
        print(e)


def convert_path(s_path):
    return s_path[s_path.find('ise'):]


def cmp_src_files(files_gs, files_gen, files_git):
    # list 2 return for the src stat values
    res = []
    print("Source code files statistics:")
    # * nr of files in git
    #   * + in IL
    #   * + in GS
    #   * + in GS and in IL
    print("%d files in git" % len(files_git))
    res.append(len(files_git))
    c_fg_gen = 0
    c_fg_gs = 0
    c_fg_gen_gs = 0
    c_fg_only = 0
    for fgit in files_git:
        # files in git and the generated links
        if fgit in files_gen:
            c_fg_gen += 1
        # files in git and the gs links
        if fgit in files_gs:
            c_fg_gs += 1
        # files in git, the generated links and the gs links
        if fgit in files_gen and fgit in files_gs:
            c_fg_gen_gs += 1
        # files only in git
        if fgit not in files_gen and fgit not in files_gs:
            #print(fgit)
            c_fg_only += 1
    print("%d files in git and in gen" % c_fg_gen)
    res.append(c_fg_gen)
    print("%d files in git and in GS" % c_fg_gs)
    res.append(c_fg_gs)
    print("%d files in git, gen and in GS" % c_fg_gen_gs)
    res.append(c_fg_gen_gs)
    print("%d files only in Git" % c_fg_only)
    res.append(c_fg_only)

    # * nr of files in IL
    #   * + in GS
    print("%d files in gen" % len(files_gen))
    res.append(len(files_gen))
    c_fgen_gs = 0
    c_fgen_only = 0
    c_fgen_not_git = 0
    for fgen in files_gen:
        if fgen in files_gs:
            c_fgen_gs +=1
        if fgen not in files_gs and fgen not in files_git:
            c_fgen_only +=1
        if fgen not in files_git:
            c_fgen_not_git +=1

    print("%d files in gen and GS" % c_fgen_gs)
    res.append(c_fgen_gs)
    print("%d files only in gen" % c_fgen_only)
    res.append(c_fgen_only)
    print("%d files in gen and not in git" % c_fgen_not_git)
    res.append(c_fgen_not_git)

    # * nr of files in GS
    print("%d files in GS" % len(files_gs))
    res.append(len(files_gs))
    c_fgs_only = 0
    for fgs in files_gs:
        if fgs not in files_gen and fgs not in files_git:
            c_fgs_only +=1
            #print(fgs)
    print("%d files only in GS" % c_fgs_only)
    res.append(c_fgs_only)
    return res


def cmp_links(gs_links, il_links, print_links=True):
    # remove duplicates they might occur due to combination of IL and IR
    unique_links = []
    for link in il_links:
        if link not in unique_links:
            unique_links.append(link)
    il_links = unique_links
    if print_links:
        print("-- Correct IL Links --")
    correct_il_links = []
    for g_link in gs_links:
        for il_link in il_links:
            # the same issue
            if g_link[0] == il_link[0]:
                # the same source code file
                if g_link[1] == il_link[1]:
                    # if not link_contained(il_link[0], il_link[1],correct_il_links):
                    if il_link not in correct_il_links:
                        if print_links:
                            print(il_link)
                        correct_il_links.append(il_link)
    if print_links:
        print("")
        print("-- Wrong IL Links --")
    wrong_il_links = []
    for w_link in il_links:
        if w_link not in correct_il_links and w_link not in wrong_il_links:
            if print_links:
                print(w_link)
            wrong_il_links.append(w_link)
    if print_links:
        print("")
        print("-- Not Found Links --")
    not_found_links = []
    for g_link in gs_links:
        if not link_contained(g_link[0], g_link[1], correct_il_links) and g_link not in not_found_links:
            # and not link_contained(g_link[0],g_link[1], not_found_links):
            # if g_link not in correct_il_links and g_link not in not_found_links:
            if print_links:
                print(g_link)
            not_found_links.append(g_link)
    if print_links:
        print("")
    # list 2 return with calculated stat values
    res = []
    print("%d IL Links " % len(il_links))
    # list 2 return with calculated stat values
    res.append(len(il_links))
    print("%d IL Links correct" % len(correct_il_links))
    res.append(len(correct_il_links))
    print("%d IL Links wrong" % len(wrong_il_links))
    res.append(len(wrong_il_links))
    print("%d Gold Standard Links" % len(gs_links))
    res.append(len(gs_links))
    print("%d Gold Standard Links not found" % len(not_found_links))
    res.append(len(not_found_links))
    if len(il_links) > 0 :
        p = len(correct_il_links)/len(il_links)
    else :
        p = 0
    r = len(correct_il_links)/len(gs_links)
    print("%0.3f Precision(IL)" % p)
    res.append(p)
    print("%0.3f Recall(IL)" % r)
    res.append(r)
    if p > 0 and r > 0:
        f_05 = (1+(0.5**2)) * ((p*r)/(((0.5**2)*p)+r))
        f_1 = (1+(1.0**2)) * ((p*r)/(((1.0**2)*p)+r))
    else:
        f_05 = 0.0
        f_1 = 0.0
    print("%0.3f F_0.5(IL)" % f_05)
    res.append(f_05)
    print("%0.3f F_1.0(IL)" % f_1)
    res.append(f_1)
    return res


def print_all_res(res):
    print("")
    print("--- CSV import : ---")
    keys = ['IL Links',
    'IL Links correct',
    'IL Links wrong',
    'Gold Standard Links',
    'Gold Standard Links not found',
    'Precision(IL)',
    'Recall(IL)',
    'F_0.5(IL)',
    'F_1.0(IL)',
    'files in git',
    'files in git and in gen',
    'files in git and in GS',
    'files in git, gen and in GS',
    'files only in Git',
    'files in gen',
    'files in gen and GS',
    'files only in gen',
    'files in gen and not in git',
    'files in GS',
    'files only in GS']
    # table header line
    head = 'Story;' + ';'.join(keys)
    print (head)
    # print the values for each story in line
    for s_res in sorted(res):
        p_line = '' + s_res + ';'
        for value in res[s_res]:
            p_line += '{}'.format(value) + ';'
        # formatting for more easy DE excel import
        if p_line.find('.') != -1:
            p_line = p_line.replace('.', ',')
        #TODO: use issue issue_id_pattern from vcs.py instead of hard coded issue id
        if p_line.find('ISE2017-') != -1:
            p_line = p_line.replace('ISE2017-', '')
        if p_line.find('ISE2016-') != -1:
            p_line = p_line.replace('ISE2016-', '')
        print(p_line)


def fetch_data_from_jira(jira, jira_project):
    # 1. Data fetching and IL log generation from interaction logs stored in Jira
    # get the issue IDs of requirments issues (task, sub-task, story, epic) which have git commits
    il_issues = jira.search_issues('project = ' + jira_project + ' and gitCommitsReferenced is not EMPTY and '
                                   'issuetype != Bug ORDER BY updated DESC',fields='key', maxResults=500)
    print("found %d issues with git commits" % len(il_issues))
    # create a timestamp based dir to store the received interaction logs in a csv file for each issue
    data_dir = getcwd() + '/data/'
    il_log_dir = data_dir  + str(datetime.now()).replace(':', '').replace(' ','_') + '/'
    create_dirs(il_log_dir)
    # generate interaction logs for issues based on recorded interaction events
    # return the list of story issues which are required for the gold standard questionnaire
    s_issues_q = gen_il_logs_raw(jira, il_log_dir + 'il/raw/', il_issues)
    # generated the interaction logs for each story
    gen_il_logs_stories(jira, il_log_dir + 'il/stories/', s_issues_q)
    convert_il_logs_stories(il_log_dir + 'il/stories/', il_log_dir + 'il/stories_cp/')
    # create link candidate questionnaires for all existing story issues
    gen_gs_quest(il_log_dir + 'gs_quest/', s_issues_q)
    return data_dir


def convert2links(files_per_issue):
    """
    convert a map with issue ids as keys and list of files as values to
    a list of links [issue_id,file]
    :param files_per_issue:  a map with issue ids as keys and list of files as values
    :return: list of links [issue_id,file]
    """
    links = []
    for issue_id in sorted(files_per_issue):
        for file in sorted(files_per_issue[issue_id]):
            link = [issue_id, file]
            if link not in links:
                links.append(link)
    return sorted(links)


if __name__ == '__main__':
    # main program to start interaction log trace link creation and comparison with links from
    # git or IR uses local data from the 'data/' folder of the project
    #
    # to use git commit specific data it is necessary to initialize the local git repository
    # with the provided 'setup_git.sh' script in the '/data/git/' directory
    #
    # for your own usage read through the comments below and change the parameters accordingly
    #
    # The modules of the program are :
    # - ir.py for IR based trace link creation
    # - vcs.py for Git Commit based trace link creation
    # - scs.py for Java soruce code structure functionality
    # for using the modules separately also see the main functions of the modules

    # if true more messages about the actual processing state are generated
    DEBUG_OUTPUT = False
    # specify if interaction log and issue data should be fetched from jira (= True)
    # or if local data should be used (= False)
    fetch_data = False
    # init the jira client
    jira = None
    # jql query to fetch all Story issues from jira_project 'ISE2017' (= project ID string in Jira)
    jira_project = 'ISE2017'
    story_q = 'project = ' + jira_project + ' and issuetype = Story ORDER BY key, updated DESC'
    try:
        # to use the jira client replace the placeholder values <..> with actual jira connection data
        jira = JIRA('<JIRA_URL>', basic_auth=('<JIRA_USR>', '<JIRA_PW>'))
    except:
        # set the jira client to None if initialisation fails
        jira = None
        pass
    # files_per_issue is the map which holds the files for each story based on the git commits diffs
    # of this story and corresponding sub-tasks
    files_per_issue = {}
    u_src_files_git = []
    # links2gen specifies which kind of links to generate for comparison with the gold standard links
    # options are :
    # * IL = links based on interaction logs
    # * IR = using either vsm or lsa
    # * GLinks = commit based links
    # * Git = also include some statistics about used source code files, Git mode is necessary for
    #         commit based link creation with GLinks, i.e. allways use 'Git|GLinks' for generating
    #         the actual commit links
    # different modes can be commbine by using the | char e.g. the used sample mode configuration
    # 'Git|IL' generates commit based links and creates the statistic about source code files usage
    links2gen = 'Git|IL'#GLinks'#Git|GLinks'#IL'IL|Git'#
    # option 1: fetch data (interaction logs, story/ sub-task text content) from jira
    if fetch_data:
        # fetch all story issues and all linked sub tasks and store the summaries and descriptions in text
        # files for each story
        s_issues = jira.search_issues(story_q, maxResults=500)
        fetch_stories(jira, s_issues)
        data_dir = fetch_data_from_jira(jira,jira_project)
    # option 2: used the already fetch data and manual extracted data from static import folder
    else:
        data_dir = getcwd() + '/data/'
    # generate a list of actual src files (= java and xml) from the local git repository
    # TODO: perform cloning first as of now a links2gen run with 'Git' is necessary in the first place
    actual_src_files = get_src_files(data_dir + 'git/')

    # generated the links of the gold standard based on the questionnaires
    gs_links = limit_links(aggregate_links(data_dir + 'gs/'))
    # create a list with issue ids from the gold standard to limit the link set to compare
    # to only contain issues which are also part of the gold standard
    gs_i_ids = get_unique_issue_id(gs_links)
    # limit the gold standard links to only contain links to files in actual_src_files
    # gs_links = limit_links2_issues(gs_i_ids, gs_links, actual_src_files)
    print("Using %d GS Links" % len(gs_links))
    print("Using %d story issues" % len(gs_i_ids))
    # list which will contain the generated links either interaction log links (IL),
    # commit links (ComL) or information retriveal (IR) link
    gen_links = []
    if links2gen.find('IL') != -1:
        # convert the raw mylyn interaction logs to a format which is comparable to the gold standard
        # and filter the IL links based on interaction log specific filter criteria if desired
        #convert_il_logs_stories(data_dir + 'raw/il/stories/', data_dir + 'gen/il/stories_cp2/')
        # use the IL_i links already created out of raw mylyn format interaction logs
        gen_links.extend(limit_links2_issues(gs_i_ids,
                                             aggregate_links(data_dir + 'il/'), actual_src_files))
        print("Found %d IL links" % len(gen_links))
    if links2gen.find('IR') != -1:
        # IR specific settings
        # the used IR algorithm either 'VSM' or 'LSA'
        ir_algo = 'VSM'
        ir_threshold = 0.2
        # path of the folder with java/ xml source code from the local git repo
        local_git_ir_repo_path = data_dir + 'git/StudentNavigator/'
        # path of jira story issues in txt files
        jira_issue_path = data_dir + 'ir/s_sts/en/'
        ir_links = limit_links2_issues(gs_i_ids, create_IR_links(local_git_ir_repo_path, jira_issue_path,
                                                                 ir_algo, ir_threshold), actual_src_files)
        print("Found %d IR links" % len(ir_links))

        # with the 'Intersect' option it is possible to combine IR and Il generated links
        if links2gen.find('Intersect') != -1:
            # intersect gen links with ir links use only the links which are in both sets of links
            isect_links = []
            for gen_link in gen_links:
                if gen_link in ir_links:
                    isect_links.append(gen_link)
            gen_links = isect_links
            print("%d links after intersection" % len(gen_links))
        else:
            # default case is add the links to list of generated links
            gen_links.extend(ir_links)
    if links2gen.find('Git') != -1:
        # use a specific revision (as revision hash) from git for commit based link creation
        # default case is None which uses the latest git revision
        git_rev_gs = None
        # generate git vcs specific information:
        git_uri = None
        local_git_dir = data_dir + 'git/'
        # try to clone a remote git repo, if there is already a git repo at local_git_dir this repo is used directly
        git_repo = clone_remote(git_uri, local_git_dir, git_rev_gs)
        files_per_issue = match_commit_diffs(git_repo, git_rev_gs, False, actual_src_files)
        # create commit based links without mapping commits from sub tasks to stories
        nsm_git_links = convert2links(files_per_issue)
        print("### Created 'not story mapped' %d Git Links" % len(nsm_git_links))
        if DEBUG_OUTPUT:
            print("List of all issue IDs found in git commits:")
            for fpi_id in sorted(files_per_issue.keys()):
                print(fpi_id)
            print("")

        u_src_files_git = get_distinct_files(files_per_issue)
        print("Found %d distinct source code files in git before alignment with actual git revision"
              % len(u_src_files_git))
        if DEBUG_OUTPUT:
            for u_src_file in u_src_files_git:
                print(u_src_file)
            print("")
        print("Initial size of generated links is %d" % len(gen_links))
        # if the jira client is initialised generate the mapping from stories to subtasks based on the query
        # for all stories in the specified jira_project
        # if not directly use the the mapping as specified in story_2_subtask_map
        if jira is not None:
            story_2_subtask_map = get_stories_st_mapping(jira.search_issues(story_q, maxResults=500))
        # XXX: only allow files which are in the gold standard
        actual_src_files = get_distinct_files_links(gs_links)
        print("Found %d distinct source code files in GS" % len(actual_src_files))
        # map sub-task files to stories, and reduce the files to only contain files in the actual git revision
        files_per_issue = trans_src_files2stories(story_2_subtask_map, files_per_issue, actual_src_files)
        u_src_files_git = get_distinct_files(files_per_issue)
        print("Found %d distinct source code files in git after alignment with actual git revision"
              % len(u_src_files_git))
        if DEBUG_OUTPUT:
            for u_src_file in u_src_files_git:
                print(u_src_file)
            print("")

        # generate links based on the files_per_issue map from git
        if links2gen.find('GLinks') != -1:
            git_links = convert2links(files_per_issue)
            print("Created %d Git Links" % len(git_links))
            gen_links.extend(limit_links2_issues(gs_i_ids, git_links, actual_src_files))
            print("Using %d Git Links for GS compare" % len(gen_links))


    if links2gen.find('SCS_R1') != -1:
        # extend the generated links with source code structure (scs) links for recall improvement
        # Option 1: add scs links before the il links are matched to the files
        #           contained in the git commits for the issues
        add_code_dep_links(gen_links, get_src_files(), 'data/scs/studentnavigator_19056e0_jdt_scs.xml', 2)
    if links2gen.find('Reduce') != -1:
        # remove all interaction links which do contain src files not in git
        gen_links = reduce_links(gen_links, files_per_issue, 'story')
        print("Reduced size of generated links is %d after matching with files from git" % len(gen_links))
        # remove all gold standard links which do contain src files not in git
        print("Initial size of GS links is %d" % len(gs_links))
        #gs_links = reduce_links(gs_links, files_per_issue, 'story')
        print("Reduced size of GS links is %d after matching with files from git" % len(gs_links))
        # TODO: move the specific configuration option to an extra operation
        # remove all xml files from links
        #gs_links = rem_links_endswith(gs_links)
        #gen_links = rem_links_endswith(gen_links)
        # remove all java files from links
        #gs_links = rem_links_endswith(gs_links,'java')
        #gen_links = rem_links_endswith(gen_links, 'java')

    # compare the gold standard links with the generated links
    if gen_links is not None and len(gen_links) > 0:
        # extend the generated links with source code structure (scs) links
        # Option 2: add scs links after the il links have been matched to the files
        #           contained in the git commits for the issues
        #add_code_dep_links(gen_links, get_src_files(), 'data/scs/studentnavigator_19056e0_jdt_scs.xml', 4)

        # story to results map for pretty printing
        stories_res = {}
        # perform a comparison for each story
        for i_id in get_unique_issue_id(gs_links):
            stories_res[i_id] = []
            story_gs_links = []
            for gs_link in gs_links:
                if gs_link[0] == i_id:
                    story_gs_links.append(gs_link)
            story_gen_links = []
            for gen_link in gen_links:
                if gen_link[0] == i_id:
                    story_gen_links.append(gen_link)
            print("-- Story %s --" % i_id)
            # story_gen_links can be a empty list that is there were no links found for a
            # particular issue, to see how many links for this issue are in the gold standard
            # it is necessary to pass the empty list to the story specific comparison and
            # handle the division by zero when calculatiing precision
            if len(story_gs_links) > 0: # and len(story_gen_links) > 0:
                # which code files are in the story:
                if DEBUG_OUTPUT:
                    print("Source code files for the story from gen links:")
                    print_distinct_files(story_gen_links)
                    print("")

                    print("Source code files for the story from GS links:")
                    print_distinct_files(story_gs_links)
                    print("")

                    print("Source code files for the story git links:")
                    if files_per_issue[i_id] is not None:
                        print_list(files_per_issue[i_id])
                    print("")
                # generate the source code files statistics:
                story_src_res = [0] * 11
                if i_id in files_per_issue.keys() and len(files_per_issue.keys()) > 0:
                    story_src_res = cmp_src_files(get_distinct_files_links(story_gs_links),
                                        get_distinct_files_links(story_gen_links),
                                        files_per_issue[i_id])

                print("")
                # perform the link comparison
                story_link_res = cmp_links(story_gs_links, story_gen_links, False)
                # add the stats to the overall results map for pretty printing
                stories_res[i_id].extend(story_link_res)
                stories_res[i_id].extend(story_src_res)
            else:
                print("Size of GS Links: %d Size of generated Links: %d" %
                      (len(story_gs_links), len(story_gen_links)))
            print("")
            print("")
        # perform a overall comparison
        print("-- 4 all %d Stories: --" % len(get_unique_issue_id(gs_links)))
        # generate the source code files statistics:
        all_src_res = [0] * 11
        if len(u_src_files_git) > 0:
            all_src_res = cmp_src_files(get_distinct_files_links(gs_links),
                            get_distinct_files_links(gen_links),
                            u_src_files_git)
        print("")
        # perform the link comparison
        all_link_res = cmp_links(gs_links, gen_links, True)
        # add the overall results to the all results map
        stories_res['all'] = []
        stories_res['all'].extend(all_link_res)
        stories_res['all'].extend(all_src_res)

        print_all_res(stories_res)
    else:
        print("No links created likely due to no operation called '%s' does exist!" % links2gen)




