"""
functions for loading data form Git
"""

from __future__ import print_function
import os
import os.path
import re
import traceback
from git import Repo
from git import RemoteProgress
from git.exc import GitCommandError
from git.exc import InvalidGitRepositoryError
from os.path import join
from os import getcwd


# ise 17/18 settings
use_file_extension = ['.java', '.xml']
use_sub_dir = ['StudentNavigator/res/', 'StudentNavigator/src/']
issue_id_pattern = 'ISE\d*-\d+'


"""
Progresse Monitor console ouutput printer for git clone/ checkout
"""
class MyProgressPrinter(RemoteProgress):
    def update(self, op_code, cur_count, max_count=None, message=''):
        print(op_code, cur_count, max_count, cur_count / (max_count or 100.0), message or "NO MESSAGE")


def clone_remote(rr_uri, l_path, rev):
    print("%s" % l_path)
    if not os.path.exists(l_path):
        os.makedirs(l_path)
        # os.makedir(l_path)
    clone_repo = None
    try:
        print("try to clone remote git repo from URI '%s' ... " % rr_uri)
        clone_repo = Repo.clone_from(rr_uri, l_path, progress=MyProgressPrinter())
    #except GitCommandError:
    except:
        #traceback.print_stack()
        # FIXME: update case 'fetch from upstream >> pull' should be handled as well here (i.e. it is not
        # necessary since we use tags from the past for our comparison
        if rr_uri is None:
            print("No git repo URI specified trying to use existing local repo @ %s" % l_path)
        else:
            print("Can NOT clone repo from URI '%s' Since there is already a repo @ %s!" % (rr_uri, l_path))
        clone_repo = None

        try:
            clone_repo = Repo(l_path)
        except InvalidGitRepositoryError:
            traceback.print_stack()
            print("No local git repo found @ %s" % l_path)
            print("Use the 'setup_git.sh' script @ '%s' to initialize the local git repository "
                  "in this directory!" % l_path)
        pass
    if rev is not None:
        gs_commit = clone_repo.commit(rev)
        clone_repo.git.checkout(gs_commit)
    return clone_repo


def include_file(file):
    if file.endswith(tuple(use_file_extension)) and file.startswith(tuple(use_sub_dir)):
        return True
    else:
        return False


def reset_issue_id (issue_id, reset_issue_id='ISE2017-32'):
    old_issue_id = issue_id
    issue_id = reset_issue_id
    print("Setting '%s' to '%s'" % (old_issue_id, issue_id))
    return issue_id


def maintain_file(file, actual_files, print_info=False):
    #return file
    # only perform a replacement if the complete path of the file in not
    # contained in the list of actual files and a java or xml file
    if file not in actual_files and file.endswith(tuple(use_file_extension)):
        for a_file in actual_files:
            # TODO: use system path api instead of string find / sub
            # get the actual name of the file from the file paths
            file_name = file[file.rfind('/') + 1:]
            if file_name in a_file:
                file = a_file
                if print_info:
                    print("Org. path '%s' and act. path '%s' contain file name '%s' org. -> act." %
                          (file, a_file, file_name))
                return file
    return file


def match_commit_diffs(git_repo, rev, process_branches=True, actual_head_files=None, print_info=False):
    if actual_head_files is None:
        # XXX: quick hack to fix keep the used internal files path structure which requires a ending '/'
        #      to restrict files correctly to the used sub folders in get_src_files()
        if not git_repo.working_dir.endswith('/'):
            actual_head_files = get_src_files(git_repo.working_dir + '/')
        else:
            actual_head_files = get_src_files(git_repo.working_dir)
    # set the last_date to the commit date of the last revision to process
    last_date = None
    if rev is not None:
        last_date = git_repo.commit(rev).committed_date

    # distinct list of processed commits to make sure a commit is only processed once
    p_commits = []
    # distinct list of commits with no jira issue ids
    p_commits_no_i = []
    # distinct list of commits with jira issue ids
    p_commits_i = []
    #list of commits used for link creation
    p_commits_used = []
    files_without_issue = []
    files_per_issue = {}
    files_without_issue_since_last_commit = []
    commits = []

    # iterate through all branches and get all commits from all branches
    for b in git_repo.references:
        #repo_heads = git_repo.head
        #print(repo_heads)
        #repo_head_names = [git_repo.heads for h in repo_heads]
        #print(repo_head_names)
        #commits = list(git_repo.iter_commits('master'))
        #if b.name == 'master':
        b_commits = list(git_repo.iter_commits(b))
        commits.extend(b_commits)
        commits = list(set(commits))
        #print("Found %d commits in '%s'" % (len(commits), b))
        print("Found %d commits in '%s' now overall %d commits" % (len(b_commits), b, len(commits)))
        # process the branches which refer to issue ids
        if process_branches:
            ref_issue_id = get_issue_id_from_ref(b)
            if ref_issue_id is not None:
                if print_info:
                    print("Processing ref issue id %s" % ref_issue_id)
                if ref_issue_id not in files_per_issue:
                    files_per_issue[ref_issue_id] = [];
                # process all commits for the branch
                # FIXME: branch commits are processed without considdering the given revision date,
                # a check for the revision date of the commits in a branch is needed!
                b_commits = list(git_repo.iter_commits(b))
                # process only the branch creating commit
                #b_commits = []
                #b_commits.append(b.commit)
                for b_commit in b_commits:
                    for file in b_commit.stats.files:
                        # FIXME: how to filter the files here reasonable ?
                        file = parse_git_mov_path(file)
                        # change the path of files from an potential early structure to the
                        # actual vcs head structure, e.g. from '/app/src/...' to '/src/...'
                        if file not in actual_head_files:
                            file = maintain_file(file, actual_head_files)
                        if file not in files_per_issue[ref_issue_id] and include_file(file):
                            files_per_issue[ref_issue_id].append(file)
                            if print_info:
                                print("Adding '%s' 4 ref issue id '%s'" % (file, ref_issue_id))

    # sort the commits according to their date
    # https://stackoverflow.com/questions/403421/how-to-sort-a-list-of-objects-based-on-an-attribute-of-the-objects
    # author
    #commits.sort(key=lambda x: x.author.email, reverse=False)
    # commit date
    #commits.sort(key=lambda x: x.committed_date, reverse=False)
    # commit date and author
    #commits.sort(key=lambda x: (x.author.email, x.committed_date), reverse=False)
    # create a map with all commits for each author
    commits_author_map = create_commit_author_map(commits)
    # process the commits for each author
    # for key in sorted(ils_map.keys(), reverse=False):
    for author in sorted(commits_author_map.keys(), reverse=False):
        for commit in commits_author_map[author]:
            # last is set so check that the actual processed commit is before this date
            if last_date is not None:
                if commit.committed_date > last_date:
                    #print("binsha  rev cmp : %s==%s" % (commit.hexsha, rev))
                    # TODO: check if time stamp can be printed as string
                    print("Actual committed_date %s is after the committed_date %s of the last commit to process" %
                          (commit.committed_date, last_date))
                    break
            if commit.hexsha not in p_commits:
                i_ids = get_issue_ids(commit.message, print_message=False)
                # if there is no issue id in a commit message a empty list is returned
                if len(i_ids) > 0:
                    for issue_id in i_ids:
                        # create a empty list for issues not already contained in the dict
                        if issue_id not in files_per_issue:
                            files_per_issue[issue_id] = [];
                        # process the files of the commit itself
                        for file in sorted(commit.stats.files):
                            # FIXME: how to filter the files here reasonable ?
                            file = parse_git_mov_path(file)
                            # change the path of files from an potential early structure to the
                            # actual vcs head structure, e.g. from '/app/src/...' to '/src/...'
                            if file not in actual_head_files:
                                file = maintain_file(file, actual_head_files)
                            if file not in files_per_issue[issue_id] and include_file(file):
                                files_per_issue[issue_id].append(file)
                                p_commits_used.append(commit)

                        # process the files from the previous commits without issue ids
                        if len(files_without_issue_since_last_commit) > 0:
                            # add the previous without issue id in commit files to the issue
                            for wi_file in sorted(files_without_issue_since_last_commit):
                                # check if file is already contained
                                if wi_file not in files_per_issue[issue_id]:
                                    files_per_issue[issue_id].append(wi_file)
                                    #print("Adding file '%s' from the %d files form the last commits without issue ID" %
                                    #     (wi_file, len(files_without_issue_since_last_commit)))
                                #else:
                                #    print("NOT Adding file '%s' from the %d files form the last commits without issue "
                                #          "ID since it is already contained" %
                                #          (wi_file, len(files_without_issue_since_last_commit)))

                    # empty the list from the previous commits without issue after all issues from the
                    # commit message have been processed (there could be more then one issue id in a commit message!)
                    #files_without_issue_since_last_commit.clear() #= []
                    #print("Clearing list of files from the last commits without issue ID containing %d files" % len(
                    #    files_without_issue_since_last_commit))
                    files_without_issue_since_last_commit = []
                    p_commits_i.append(commit.hexsha)
                # if the list of items from the iterator is empty no issue id has been found in the commit message
                # the files from these commits have to be added to next commit with an issue id
                else:
                    if print_info:
                        print("No jira issue id in commit %s" % commit)
                    for file in sorted(commit.stats.files):
                        file = parse_git_mov_path(file)
                        # change the path of files from an potential early structure to the
                        # actual vcs head structure, e.g. from '/app/src/...' to '/src/...'
                        if file not in actual_head_files:
                            file = maintain_file(file, actual_head_files)
                        # process the collected files
                        if file not in files_without_issue_since_last_commit and include_file(file):
                            files_without_issue_since_last_commit.append(file)
                            p_commits_used.append(commit)
                        # also create list with all files not in the commits with issue ids for statistic purpose
                        if file not in files_without_issue:
                            files_without_issue.append(file)
                    p_commits_no_i.append(commit.hexsha)
                p_commits.append(commit.hexsha)

    print("Processed %d commits form %d branches" % (len(p_commits), len(git_repo.references)))
    print("Used files from %d commits for link creation" % len(list(set(p_commits_used))))
    print("%d commits did contain an jira issue ID" % len(p_commits_i))
    print("%d commits did NOT contain an jira issue ID" % len(p_commits_no_i))
    print("Found %d files from commit diffs without issues: " % len(files_without_issue))
    for file in files_without_issue:
        if not include_file(file):
            # remove all not matching files
            files_without_issue = list(filter(lambda a: a != file, files_without_issue))

    print("%d files after filer: " % len(files_without_issue))
    if print_info:
        files_with_issue = get_distinct_files(files_per_issue)
        print("Files only contained in commits without issue ids:")
        c_fwfw = 0
        for fwi in files_without_issue:
            if fwi in files_with_issue:
                c_fwfw +=1
            else:
                print(fwi)
        print("%d files are contained in the commits with and without issue ids" % c_fwfw)

    return files_per_issue


def parse_git_mov_path(path, debug=False):
    # make shure we have a move
    if '{' and '=>' and '}' in path:
        # extract the base path
        base_path = path[:path.find('{')]
        # the performed move operation
        mov_op = path[path.find('{')+1:path.find('}')]
        # the old file from the move operation
        old_file_path = mov_op[:mov_op.find('=>')]
        # the new file from the move operation
        new_file_path = mov_op[mov_op.find('=>')+3:]
        # recreate the file paths
        new_path = base_path + new_file_path
        old_path = base_path + old_file_path
        # debug print for check
        if debug:
            print("old path: %s" % old_path)
            print("new path: %s" % new_path)
        # return the new path as path
        path = new_path
    return path


def get_distinct_files(files_per_issue):
    u_src_files = []
    for issue in files_per_issue:
        for file in files_per_issue[issue]:
            if file not in u_src_files:
                u_src_files.append(file)
    return list(set(u_src_files))


def get_issue_id_from_ref(ref):
    """
    pare the name of a git remote reference for a jira issue id and return the issue id if one is contained
    :param ref:
    :return:
    """
    bname_is = re.finditer(issue_id_pattern, ref.name)
    bname_is = list(bname_is)
    if len(bname_is) > 0:
        for bname_i in bname_is:
            issue_id = bname_i.group(0)
            return issue_id
    return None


def get_src_files(base_path='/path/2/loacal/git/repo/',
                  use_file_extension_p=use_file_extension,
                  use_sub_dir_p=use_sub_dir):
    base_path = base_path.replace("\\", '/')
    print("New base_path is now: %s" % base_path)
    src_set = []
    for path, subdirs, files in os.walk(base_path):
        for name in files:
            p_path = os.path.join(path, name).replace("\\", '/')
            # only add src files which are in the defined sub dirs and are files of type as in use file extension
            for fe in use_file_extension_p:
                for sd in use_sub_dir_p:
                    if p_path.endswith(fe) and p_path.find(sd) != -1:
                        # use the relative path in the git repo dir by removing the base path
                        src_set.append(p_path.replace(base_path, ''))
    print("Using %d source code files" % len(src_set))
    return sorted(src_set)


def get_all_commits(git_repo):
    commits = []
    for b in git_repo.references:
        b_commits = list(git_repo.iter_commits(b))
        commits.extend(b_commits)
        commits = list(set(commits))
        print("Found %d commits in '%s' now overall %d commits" % (len(b_commits), b, len(commits)))
    # sort the list of all commits by commited date
    commits.sort(key=lambda x: x.committed_date, reverse=False)
    return commits


def get_issue_ids(commit_message, print_message=True):
    #print(commit_message)
    i_ids = []
    ms = re.finditer(issue_id_pattern, commit_message)
    # finditer() return a interator with 0 elements if nothing is found
    if ms is not None:
        ms = list(ms)
        #FIXME: somehow the call to sort() the ms list stopped working !?
        # if the list of items from the iterator contains s.th. a issue id has been found
        if len(ms) > 0:
            #ms = sorted(ms)
            for m in ms:
                issue_id = m.group(0)
                if print_message:
                    print("Commit message '%s' contains Jira issue id %s" %
                          (commit_message.replace('\n', ' '), issue_id))
                if issue_id not in i_ids:
                    i_ids.append(issue_id)
    return i_ids


def create_commit_author_map(commits):
    commits_auth_map = {}
    for commit in commits:
        if commit.author.name not in commits_auth_map:
            commits_auth_map[commit.author.name] = [commit]
        else:
            commits_auth_map[commit.author.name].append(commit)
    # sort the commits for each author by date
    for author in commits_auth_map:
        commits_auth_map[author].sort(key=lambda x: x.committed_date, reverse=False)
    return commits_auth_map


if __name__ == '__main__':
    # test the git commit parsing functionality with the provided local git repo

    # since a local git repo is used no git_uri is required
    git_uri= None
    # the easiest way to specify user and password protected remote git repository is
    # to directly provide user and passwordin the git_uri like:
    # git_uri = 'https://<USER>:<PASSWORD>@git.repo.org/my/remote/git/repo'
    # e.g. :
    # git_uri = 'https://developer:1111_42!pw@git.eclipse.org/still/fictional/repo'

    # local directory in which a potential remote git repo will be cloned
    # if the directory already exist it is directly used as git repo, like in this sample case
    local_git_dir = getcwd() + '/data/git/'
    git_repo = clone_remote(git_uri, local_git_dir, None)

    # test processing references (branch names) for issue id
    for ref in git_repo.references:
        ref_i_id = get_issue_id_from_ref(ref)
        if ref_i_id is not None:
            print("Ref '%s' contains issue id '%s'" % (ref, ref_i_id))

    # test to get issue ids and files from commits
    files_per_issue = match_commit_diffs(git_repo, None)
    print(files_per_issue)
    u_src_files = get_distinct_files(files_per_issue)
    print()
    print("Found %d distinct issue IDs and %d distinct source code files in commits"
          % (len(files_per_issue.keys()), len(u_src_files)))
    print("Issue Ids to src file mapping:")
    for issue in sorted(files_per_issue):
        for file in sorted(files_per_issue[issue]):
            print("%s:%s" % (issue, file))




